import logging
import os

from piped import service, util, exceptions
from twisted.internet import defer
from zope import interface

import psutil

from piped_statsd import client


logger = logging.getLogger(__name__)


class MetricReporter(object):
    """ Metric reporter that is invoked periodically by the MetricsReporter-service.

    Metrics should be prefixed with `prefix` and tagged with `tags`. The clients are
    available through "statsd".
    """

    def __init__(self, prefix, statsd, client_name='default', tags=None):
        self.prefix = prefix
        self.statsd = getattr(statsd, client_name)
        self.tags = tags

    def __call__(self):
        raise NotImplementedError()


class ProcessInfoReporter(MetricReporter):

    def __init__(self, *a, **kw):
        self.values = set(kw.pop('values', []))
        super(ProcessInfoReporter, self).__init__(*a, **kw)
        self.process = psutil.Process(os.getpid())

        for value in self.values:
            if not hasattr(self, 'report_' + value):
                raise exceptions.ConfigurationError('unknown process value: [{0}]'.format(value))

    def __call__(self):
        for value in self.values:
            getattr(self, 'report_' + value)()

    def report_connections(self):
        value = len(self.process.get_connections())
        self.statsd.gauge(self.prefix + '.connections', value, tags=self.tags)

    def report_open_files(self):
        value = len(self.process.get_open_files())
        self.statsd.gauge(self.prefix + '.open_files', value, tags=self.tags)

    def report_memory(self):
        memory_info = self.process.get_memory_info()
        self.statsd.gauge(self.prefix + '.rss', memory_info.rss, tags=self.tags)
        self.statsd.gauge(self.prefix + '.vms', memory_info.vms, tags=self.tags)

    def report_cpu_percent(self):
        self.statsd.gauge(self.prefix + '.cpu_percent', self.process.get_cpu_percent(), tags=self.tags)

    def report_threads(self):
        self.statsd.gauge(self.prefix + '.threads', len(self.process.get_threads()), tags=self.tags)


class MetricsReporter(service.PipedService):
    """ Service that periodically lets reporters send their gauges to statsd.

    Assumes the service configuration has "metrics.enabled" being true.

    Metrics are prefixed with "metrics.prefix" every "metrics.interval" seconds
    (defaults to 5) and with "metrics.tags" as tags.

    "metrics.reporters" should be a list of reporter-configurations. The only required
    value in a reporter-configuration is "reporter", which is the name of the reporter.
    Any other values are passed to the reporter. "client_name" can be set to use a
    specific statsd-client.

    Available reporters:

        - "process". Takes "values" as a list of values to report, with possible values
        being "cpu_percent", "connections", "open_files", "memory" and "threads".

    """
    interface.classProvides(service.IPipedService)

    _standard_reporters = dict(
        process=ProcessInfoReporter
    )

    def configure(self, runtime_environment):
        service.PipedService.configure(self, runtime_environment)
        if not self.is_enabled():
            return

        metric_configuration = runtime_environment.get_configuration_value('metrics', dict())

        self.prefix = metric_configuration.get('prefix', '')
        self.tags = metric_configuration.get('tags', [])
        self.interval = metric_configuration.get('interval', 5)

        self.statsd = client.StatsdManager()
        self.statsd.configure(runtime_environment)

        self._reporters = []

        for reporter_configuration in metric_configuration.get('reporters', list()):
            if isinstance(reporter_configuration, dict):
                Reporter = self._standard_reporters.get(reporter_configuration.pop('reporter'))
            else:
                Reporter = self._standard_reporters.get(reporter_configuration)
                reporter_configuration = dict()

            try:
                self.add_reporter(Reporter(self.prefix, self.statsd, tags=self.tags, **reporter_configuration))
            except Exception:
                logger.exception('Error making reporter')
                raise exceptions.ConfigurationError('invalid reporter configuration: [{0!r}]'.format(reporter_configuration))

    def is_enabled(self):
        return self.runtime_environment.get_configuration_value('metrics.enabled', False)

    @defer.inlineCallbacks
    def run(self):
        while self.running and self.is_enabled():
            for reporter in self._reporters:
                try:
                    reporter()
                except Exception as e:
                    logger.exception('unhandled exception for reporter: [{0!r}]'.format(reporter))

            yield util.wait(self.interval)

    def add_reporter(self, reporter):
        self._reporters.append(reporter)



import time

import statsd


class PipedStatsd(statsd.DogStatsd):

    def time_deferred(self, d, metric, tags, sample_rate=1):
        def _time_and_passthrough(result, start):
            self.timing(metric, time.time() - start, tags=tags, sample_rate=sample_rate)
            return result

        d.addBoth(_time_and_passthrough, time.time())
        return d


class StatsdManager(object):
    """ Returns `PipedStatsd`-clients as attributes.

    Assuming statsd-configurations under "statsd" map logical names to `PipedStatsd`-configurations, which
    currently only takes "host" and "port" as options, the manager will return `PipedStatsd`-clients with the
    corresponding configuration.

    The clients are accessible as manager_instance.logical_name.

    If a client does not have a configuration, it defaults to `host="localhost"` and "port=8125".

    Also provides the statsd-functions of the "default"-client directly. i.e. `manager_instance.gauge(...)` will
    use the "default"-client, however it is configured.
    """

    _statsd_keywords = set(["decrement", "gauge", "histogram", "increment", "set", "timed", "time_deferred", "timing"])

    def __init__(self):
        self._client_by_name = dict()

    def configure(self, runtime_environment):
        self.runtime_environment = runtime_environment
        self.default_config = runtime_environment.get_configuration_value('statsd.default', dict(host='localhost', port=8125))
        self.default = PipedStatsd(**self.default_config)

    def __getattr__(self, item):
        if item in self._statsd_keywords:
            return getattr(self.default, item)

        if item not in self._client_by_name:
            config = self.runtime_environment.get_configuration_value('statsd.{0}'.format(item), Ellipsis)
            if config is not Ellipsis:
                self._client_by_name[item] = PipedStatsd(**config)
            else:
                self._client_by_name[item] = self.default

        return self._client_by_name[item]



# Copyright (c) 2013, Found A/S and Piped Project Contributors.
# See LICENSE for details.
import os
import sys

from setuptools import setup, find_packages

# When pip installs anything from packages, py_modules, or ext_modules that
# includes a twistd plugin (which are installed to twisted/plugins/),
# setuptools/distribute writes a Package.egg-info/top_level.txt that includes
# "twisted".  If you later uninstall Package with `pip uninstall Package`,
# pip removes all of twisted/ instead of just Package's twistd plugins.  See
# https://github.com/pypa/pip/issues/355
#
# To work around this problem, we monkeypatch
# setuptools.command.egg_info.write_toplevel_names to not write the line
# "twisted".  This fixes the behavior of `pip uninstall Package`.  Note that
# even with this workaround, `pip uninstall Package` still correctly uninstalls
# Package's twistd plugins from twisted/plugins/, since pip also uses
# Package.egg-info/installed-files.txt to determine what to uninstall,
# and the paths to the plugin files are indeed listed in installed-files.txt.
try:
    from setuptools.command import egg_info
    egg_info.write_toplevel_names
except (ImportError, AttributeError):
    pass
else:
    def _top_level_package(name):
        return name.split('.', 1)[0]

    def _hacked_write_toplevel_names(cmd, basename, filename):
        pkgs = dict.fromkeys(
            [_top_level_package(k)
                for k in cmd.distribution.iter_distribution_names()
                if _top_level_package(k) != 'piped'
            ]
        )
        cmd.write_file("top-level names", filename, '\n'.join(pkgs) + '\n')

    egg_info.write_toplevel_names = _hacked_write_toplevel_names


here = os.path.abspath(os.path.dirname(__file__))

# add ourselves to the package path so we can get the version from the source tree
sys.path.insert(0, here)
import piped_statsd

setup(
    name = 'piped.contrib.statsd',
    license = 'MIT',

    author = 'Piped Project Contributors',
    author_email = 'piped@librelist.com',
    url = 'http://piped.io',

    packages = find_packages(where=here) + ['piped.plugins'],

    version = str(piped_statsd.version),
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Framework :: Twisted',
        'Operating System :: OS Independent',
    ],
    description = 'StatsD-utilities for Piped.',

    install_requires = ['piped', 'dogstatsd-python', 'psutil', 'setuptools']
)